#!/usr/bin/env python
#
# ---------------------------------------------------------------------------------
# Examples from the MCMC seminar
#       METROPOLIS-HASTINGS ALGORITHM: look at the end for the main program.
#       this file runs a simple metropolis method on a function f(x)
#       and displays the function _and_ the histogram calculated from the MCMC run. 
# ---------------------------------------------------------------------------------
# The MIT License (MIT)
# Copyright (c) 2013 Peter Beerli
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this 
# software and associated documentation files (the "Software"), to deal in the Software 
# without restriction, including without limitation the rights to use, copy, modify, 
# merge, publish, distribute, sublicense, and/or sell copies of the Software, and to 
# permit persons to whom the Software is furnished to do so, subject to the 
# following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies 
# or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
# PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
# OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
#
# Metropolis algorithm
#
#
from math import *
from collections import Counter
import random 
import matplotlib
import matplotlib.pyplot as plt
#
##########################################################
#
# function to integrate, this function has three humps that become less
# peaked from left to right.
def f(x):
    return exp(-x**2/10.) + 0.5 * exp(-(x-20.)**2/20.) + 0.2 * exp(-(x-40.)**2/200.)
#


########################################## 
# helper functions
#
# generates an array of float values between start and stop using a stepsize
def seq(start, stop, step=0.1):
    n = int(round((stop - start)/float(step)))
    return([start + step*i for i in range(n+1)])

# plots the function and the histogram from the MCMC
# the function and the histogram are scaled that the area under the curve is 1.0
# this could be improved: the area under the original function is ~15.
def plot(record, f,domain,):
    stepsize = (float(domain[1])-domain[0])/300.
    x = seq(domain[0],domain[1],stepsize)
    y = [f(v) for v in x]
    a = sorted(zip(x,y))
    x,y = zip(*a)
    sum = 0.0
    for yi in y:
        sum += yi * stepsize
    y2 = []
    for yi in y:
        y2.append(yi / sum)
    n, bins, patches = plt.hist(record, 100, normed=True, histtype='stepfilled')
    plt.setp(patches, 'facecolor', 'g', 'alpha', 0.75)
    plt.plot(x,y2,'r-')
    plt.show()

############################################################
#
# metropolis-hastings method
#
# proposal function
def equal_proposal(x,delta,domain):
    dx = delta * random.uniform(-0.5,0.5)
    if domain[0] < x+dx < domain[1]:
        return x+dx
    else:
        proposal(x,delta,domain)

def unequal_proposal(x,delta,domain):
    low = max(x - 2.0*delta,domain[0])
    high = min(x + delta,domain[1])
    newx = random.triangular(low, high,x)
    while not (domain[0] < newx < domain[1]):
        newx = random.triangular(low, high,x)
    return newx

def pdf_triangular(a,b,c,x): #low, high, mode, x
    if a <= x <=b:
        if x <= c:
            return 2.0*(x-a)/((b-a)*(c-a))
        else:
            return 2.0*(b-x)/((b-a)*(b-c))
    else:
        return 0.0

def hastings_ratio(oldx,newx,delta, domain):
    olow = max(oldx - 2.0 *delta,domain[0])
    ohigh = min(oldx + delta,domain[1])
    nlow = max(newx - 2.0* delta,domain[0])
    nhigh = min(newx + delta,domain[1])
    q2 = pdf_triangular(olow,ohigh,oldx,newx)
    q1 = pdf_triangular(nlow,nhigh,newx,oldx)
    #print q1,q2,olow,oldx, ohigh,nlow,newx, nhigh
    return q1/q2

# acceptance rejection step
def metropolis_hastings_accept(newx,oldx,delta,domain):
    nom = f(newx)
    denom = f(oldx)
    r = random.uniform(0.0,1.0)
    h = hastings_ratio(oldx,newx,delta,domain)
    if mh:
        test = r < nom/denom * h
    else:
        test = r < nom/denom
    if test:
        accept='* '
    else:
        accept='  '
    print accept, r, "<" , nom/denom,"*", h, "=", nom/denom * h
    if test:
        return [True,newx]
    else:
        return [False,oldx]

# mcmc
def mcmc(n=1000, delta=0.1, domain=[-100,100],startval = 0.0):
    oldx = startval
    accept = []
    record = []
    for ni in range(n):
        newx = unequal_proposal(oldx,delta,domain)
        a,b = metropolis_hastings_accept(newx,oldx, delta, domain)
        accept.append(a)
        record.append(b)
        oldx = b
    aa = Counter(accept)
    return [aa[True]/float(n),record]

##########################################################
#
# main loop
#
# set one or the other of the mh condition, the first one delivers false results 
#mh=False #this sets to metropolis using the triangular proposal distribution
mh=True  #this sets to metropolis-HASTINGS using the triangular proposal distribution
domain = [-30,100]
# number of runs n=100000, delta=10.0, the domain is over -30 to 100, and the
# start value is -20
accept, record = mcmc(100000,20.0,domain,-20)
# show the results
print "Acceptance ratio", accept
plot(record, f,domain)
